package com.perforce.hws.testing.deployment.deb;

import com.perforce.hws.automation.OsHelpers;
import com.perforce.hws.automation.ProjectLocations;
import com.perforce.hws.testing.deployment.DeploymentConventions;
import com.perforce.hws.util.ProcessHelpers;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

/**
 * Methods common to deploying .deb files.
 */
public interface DebDeployment extends
        DeploymentConventions,
        OsHelpers,
        ProcessHelpers,
        ProjectLocations {

    Logger logger = LoggerFactory.getLogger(DebDeployment.class);

    default void installOldDebFile(String debFile) {
        logger.info("installing base package from {}", debFile);

        try {
            logger.info(execOutOrFail("dpkg", "-i", debFile));
            logger.info(execOutOrFail("/opt/perforce/helix-web-services/sbin/configure-helix-web-services"));
            logger.info(execOutOrFail("service", "helix_web_services", "start"));
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }


    default void installDebFile(String debFile) {
        logger.info("installing base package from {}", debFile);

        try {
            logger.info(execOutOrFail("dpkg", "-i", debFile));
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Uses project conventions to execute an upgrade from files in one of two locations:
     *
     * 1. The EC path:
     *     //depot/[codeline]/p4-bin/bin.ubuntu[12,14]x86_64/helix-web-services-deb-pretest.tgz
     *
     * 2. The local build directory (for local testing):
     *     ./install4j/build/helix-web-services_[version]_amd64.deb
     *
     * We look for the EC path first, otherwise, assume we're using the local build.
     */
    default void upgradeDebBuild() {
        try {
            String debFile = obtainDebBuild();

            if (debFile == null) {
                throw new IllegalStateException("Unable to locate local .deb package");
            }

            logger.info("upgrading package from {}", debFile);
            logger.info(execOutOrFail("dpkg", "--force-confold", "-i", debFile));
            logger.info(execOutOrFail("service", "helix-ws", "stop"));
            logger.info(execOutOrFail("service", "helix-ws", "start"));

        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    default String obtainDebBuild() {
        String pretestArchive = obtainDebPretestArchive();
        logger.info("pretestArchive {}", pretestArchive);
        if (pretestArchive == null) {
            return localDebBuildPath();
        }
        return pretestArchive;
    }

    default String localDebBuildPath() {
        try {
            Path install4jBuild = Paths.get(projectRootOrFail(), "install4j", "build");
            Path pkg = Files.newDirectoryStream(install4jBuild, "helix-web-services*.deb").iterator().next();
            return pkg.toAbsolutePath().toString();
        } catch (IOException io) {
            throw new UncheckedIOException(io);
        }
    }

    default String obtainDebPretestArchive() {
        try {
            Path archivePath = null;
            if (isUbuntu()) {
                archivePath = Paths.get(projectRootOrFail(),  "..", "..", "p4-bin",
                        "bin.ubuntu12x86_64", "helix-web-services-deb-pretest.tgz");
            } else {
                throw new IllegalStateException("OS not supported");
            }

            logger.info("looking for archive in {}", archivePath.toAbsolutePath().toString());

            if (archivePath != null && Files.exists(archivePath)) {
                return unpackDebArchive(archivePath.toAbsolutePath().toString());
            }
            return null;
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Unpacks the archive file (in a local location) and return the local .deb file name.
     *
     * @param installerArchive
     * @return
     */
    default String unpackDebArchive(String installerArchive, String localDirName) {
        try {
            Files.createDirectories(Paths.get(localDirName));

            if (isVMWareFolder()) {
                logger.info("vmware folder tar: {}", execOutOrFail("sudo", "-u", vmwareUser(), "tar", "-x", "-z", "-p", "-C", localDirName, "-f", installerArchive));
            } else {
                logger.info("tar: {}", execOutOrFail("tar", "-x", "-z", "-C", localDirName, "-f", installerArchive));
            }

            Path pkg = Files.newDirectoryStream(Paths.get(localDirName), "helix-web-services*.deb").iterator().next();
            return pkg.toAbsolutePath().toString();

        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    default String unpackDebArchive(String installerArchive) {
        return unpackDebArchive(installerArchive, "tmp_installer_archive");
    }
}
