package com.perforce.hws.testing.reporting;

import com.esotericsoftware.yamlbeans.YamlWriter;
import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.core.Filter;
import org.apache.logging.log4j.core.Layout;
import org.apache.logging.log4j.core.LoggerContext;
import org.apache.logging.log4j.core.appender.AbstractOutputStreamAppender;
import org.apache.logging.log4j.core.appender.OutputStreamManager;
import org.apache.logging.log4j.core.config.Configuration;
import org.apache.logging.log4j.core.config.LoggerConfig;
import org.apache.logging.log4j.core.layout.PatternLayout;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.testng.*;
import org.testng.xml.XmlSuite;

import java.io.ByteArrayOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

/**
 * Created by tristan on 5/10/16.
 */
public class YamlReporter implements IReporter, IInvokedMethodListener2, ITestListener {

    static final Logger logger = LoggerFactory.getLogger(YamlReporter.class);

    StringAppender stringAppender = StringAppender.createStringAppender(null);

    List<MethodResult> methodResults = new ArrayList<>();

    public YamlReporter() {
        System.out.println("Configuring log4j");
        final LoggerContext ctx = (LoggerContext) LogManager.getContext(false);
        final Configuration config = ctx.getConfiguration();
        stringAppender.start();
        config.getRootLogger().setLevel(Level.ALL);
        stringAppender.addToLogger("root", Level.ALL);
    }


    @Override
    public void beforeInvocation(IInvokedMethod method, ITestResult testResult) {
        logger.error("beforeInvocation (old) called");
        throw new IllegalStateException("not expected to be used");
    }

    @Override
    public void afterInvocation(IInvokedMethod method, ITestResult testResult) {
        logger.error("afterInvocation (old) called");
        throw new IllegalStateException("not expected to be used");
    }

    @Override
    public void beforeInvocation(IInvokedMethod method, ITestResult testResult, ITestContext context) {
        stringAppender.reset();
    }

    @Override
    public void afterInvocation(IInvokedMethod method, ITestResult testResult, ITestContext context) {
        MethodResult methodResult = new MethodResult();
        methodResult.setClassName(method.getTestMethod().getTestClass().getName());
        methodResult.setName(method.getTestMethod().getConstructorOrMethod().getName());

        if (testResult.isSuccess()) {
            methodResult.setResultType(MethodResult.ResultType.PASSED);
        } else if (testResult.getStatus() == ITestResult.FAILURE) {
            methodResult.setResultType(MethodResult.ResultType.FAILED);
        } else if (testResult.getStatus() == ITestResult.SKIP){
            methodResult.setResultType(MethodResult.ResultType.SKIPPED);
        } else {
            methodResult.setResultType(MethodResult.ResultType.FAILED);
        }

        methodResult.setSuite(context.getSuite().getName());
        methodResult.setLog(stringAppender.getOutput());
        methodResults.add(methodResult);

        logger.info("afterInvocation methodResult {}", methodResult);
        if (testResult.getThrowable() != null) {
            logger.error("afterInvocation throwable", testResult.getThrowable());
        }
    }

    @Override
    public void generateReport(List<XmlSuite> xmlSuites, List<ISuite> suites, String outputDirectory) {

        try {
            Path resultsFile = Paths.get(outputDirectory, "results.yaml").toAbsolutePath();

            if (!Files.exists(resultsFile.getParent())) {
                Files.createDirectories(resultsFile.getParent());
            }

            logger.info("generating {}", resultsFile.toString());
            YamlWriter yamlWriter = new YamlWriter(new FileWriter(resultsFile.toFile()));
            yamlWriter.getConfig().setClassTag("result", MethodResult.class);
            yamlWriter.write(methodResults);
            yamlWriter.close();

        } catch (IOException e) {
            logger.error("Unable to create test results", e);
        }
    }

    /**
     * This creates
     */
    public static class StringAppender extends AbstractOutputStreamAppender<StringAppender.StringOutputStreamManager> {
        static LoggerContext context = (LoggerContext) LogManager.getContext(false);
        static Configuration configuration = context.getConfiguration();
        StringOutputStreamManager manager;

        private StringAppender(String name, Layout<? extends Serializable> layout, Filter filter, StringOutputStreamManager manager, boolean ignoreExceptions, boolean immediateFlush) {
            super(name, layout, filter, ignoreExceptions, immediateFlush, manager);
            this.manager = manager;
        }

        public static StringAppender createStringAppender(String nullablePatternString) {
            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            PatternLayout layout;
            if (nullablePatternString == null) {
                layout = PatternLayout.createDefaultLayout();
            } else {
                layout = PatternLayout.createLayout(nullablePatternString, null, configuration, null, null, true, false, null, null);
            }

            return new StringAppender(
                    "StringAppender",
                    layout,
                    null,
                    new StringOutputStreamManager(outputStream, "StringStream", layout, false),
                    false,
                    true);
        }

        public void addToLogger(String loggerName, Level level) {
            LoggerConfig loggerConfig = configuration.getLoggerConfig(loggerName);
            loggerConfig.addAppender(this, level, null);
            context.updateLoggers();
        }

        public void removeFromLogger(String loggerName) {
            LoggerConfig loggerConfig = configuration.getLoggerConfig(loggerName);
            loggerConfig.removeAppender("StringAppender");
            context.updateLoggers();
        }

        public String getOutput() {
            manager.flush();
            return new String(manager.getStream().toByteArray());
        }

        public void reset() {
            manager.reset();
        }

        /**
         * StringOutputStreamManager to manage an in memory bytestream representing our stream
         */
        public static class StringOutputStreamManager extends OutputStreamManager {
            ByteArrayOutputStream stream;

            protected StringOutputStreamManager(ByteArrayOutputStream os, String streamName, Layout<?> layout, boolean writeHeader) {
                super(os, streamName, layout, writeHeader);
                stream = os;
            }

            public ByteArrayOutputStream getStream() {
                return stream;
            }

            public void reset() {
                stream.reset();
            }
        }
    }

    @Override
    public void onFinish(ITestContext context) {
        logger.info("finish");
    }

    @Override
    public void onTestStart(ITestResult result) {
        logger.info("testStart {}", result.getName());
    }

    @Override
    public void onTestSuccess(ITestResult result) {
        logger.info("testSuccess {}", result.getName());
    }

    @Override
    public void onTestFailure(ITestResult result) {
        logger.error("testFailure {}", result.getName());
        if (result.getThrowable() != null) {
            logger.error("exception", result.getThrowable());
        }
    }

    @Override
    public void onTestSkipped(ITestResult result) {
        logger.info("testSkipped {}", result.getName());
    }

    @Override
    public void onTestFailedButWithinSuccessPercentage(ITestResult result) {
        logger.info("onTestFailedButWithinSuccessPercentage {}", result.getName());
    }

    @Override
    public void onStart(ITestContext context) {
        logger.info("start");
    }
}
