package com.perforce.hws.testing.server;

import com.perforce.hws.testing.ClientProvider;
import com.perforce.hws.util.StreamUtils;
import com.perforce.hwsclient.ApiClient;
import com.perforce.hwsclient.DefaultApi;
import com.perforce.hwsclient.models.LoginResponse;

import org.testng.Assert;
import org.testng.annotations.Test;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.X509Certificate;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

/**
 * Basic assertions around authorization behavior in Helix Web Services.
 * <p>
 * HWS has two kinds of authentication sources: "project" servers, which on
 * Helix Cloud are actually handled outside of p4d, and "onsite" which is
 * just handled via p4d. There are two interfaces, since Helix Cloud can need
 * to obtain the project auth *and* the p4d auth separately for each project.
 */
public class AuthTests implements StreamUtils {

    /**
     * Minimal "p4" authentication test. The client is created by our
     * typical data provider, which is used in very many test methods.
     *
     * @param api the api
     */
    @Test(dataProvider = "ProjectClientAsJDoe",
    	  dataProviderClass = ClientProvider.class)
    public void p4ClientGeneratesTicket(final DefaultApi api) {
        assertNotNull(api);
    }

    /**
     * Illegal auth throws exception.
     *
     * @throws IOException Signals that an I/O exception has occurred.
     */
    @Test(expectedExceptions = RuntimeException.class)
    public void illegalAuthThrowsException() throws IOException {
        ClientProvider.createP4Client("jdoe", "invalid");
    }

    /**
     * Minimal "project" authentication test. The client is created by our
     * typical data provider, which is used in very many test methods.
     *
     * @param api the api
     */
    @Test(dataProvider = "ProjectClientAsJDoe",
    		dataProviderClass = ClientProvider.class)
    public void projectClientGeneratesTicket(final DefaultApi api) {
        assertNotNull(api);
    }

    /**
     * Test server login.
     */
    @Test
    public void testServerLogin() {
    	ApiClient client = new ApiClient(
    			true, ClientProvider.getConfig().getLocalHws());
    	client.setCredentials(ClientProvider.DEFAULT_USER, 
    						  ClientProvider.DEFAULT_PASSWORD);
    	DefaultApi api = client.createDefaultService();
    	assertNotNull(api.serverLogin("localhost").getTicket());
    }
    
    /**
     * Empty auth throws400 error.
     */
    @Test
    public void emptyAuthThrows403Error() {
        try {
            // Create a trust manager that does not validate certificate chains
            TrustManager[] trustAllCerts =
            	new TrustManager[] {new X509TrustManager() {
                public java.security.cert.X509Certificate[] getAcceptedIssuers() {
                    return null;
                }
                public void checkClientTrusted(final X509Certificate[] certs,
                							   final String authType) {
                }
                public void checkServerTrusted(final X509Certificate[] certs,
                							   final String authType) {
                }
            }
            };

            // Install the all-trusting trust manager
            SSLContext sc = SSLContext.getInstance("SSL");
            sc.init(null, trustAllCerts, new java.security.SecureRandom());
            HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());

            // Create all-trusting host name verifier
            HostnameVerifier allHostsValid = (hostname, session) -> true;

            // Install the all-trusting host verifier
            HttpsURLConnection.setDefaultHostnameVerifier(allHostsValid);

            URL url = new URL(ClientProvider.getConfig().getLocalHws()
            		+ "/api/hws/v16.1/login");

            HttpURLConnection conn = (HttpURLConnection) url.openConnection();

            conn.setDoOutput(true);
            conn.setRequestMethod("POST");
            conn.setRequestProperty("Content-Type", "application/json");
            conn.setRequestProperty("Content-Length", "0");

            try (OutputStream os = conn.getOutputStream()) {
            }

            try (InputStream inputStream = conn.getInputStream()) {
            } catch (IOException e) {
                assertTrue(e.getMessage().startsWith(
                		"Server returned HTTP response code: 403"));
                return;
            }

            Assert.fail("Exception expected");
        } catch (IOException | NoSuchAlgorithmException | KeyManagementException e) {
            Assert.fail("Unexpected problem", e);
        }
    }


    /**
     * Illegal json throws400 error.
     */
    @Test
    public void illegalJsonThrows403Error() {
        try {
            // Create a trust manager that does not validate certificate chains
            TrustManager[] trustAllCerts = 
            	new TrustManager[] {new X509TrustManager() {
                public java.security.cert.X509Certificate[] getAcceptedIssuers() {
                    return null;
                }
                public void checkClientTrusted(final X509Certificate[] certs, 
                							   final String authType) {
                }
                public void checkServerTrusted(final X509Certificate[] certs,
                						       final String authType) {
                }
            }
            };

            // Install the all-trusting trust manager
            SSLContext sc = SSLContext.getInstance("SSL");
            sc.init(null, trustAllCerts, new java.security.SecureRandom());
            HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());

            // Create all-trusting host name verifier
            HostnameVerifier allHostsValid = (hostname, session) -> true;

            // Install the all-trusting host verifier
            HttpsURLConnection.setDefaultHostnameVerifier(allHostsValid);

            URL url = new URL(ClientProvider.getConfig().getLocalHws()
            		+ "/api/hws/v16.1/login");

            HttpURLConnection conn = (HttpURLConnection) url.openConnection();

            String badJson = "{ \"user\": \"huh\" ";

            conn.setDoOutput(true);
            conn.setRequestMethod("POST");
            conn.setRequestProperty("Content-Type", "application/json");
            conn.setRequestProperty("Content-Length",
            		Integer.toString(badJson.length()));

            try (OutputStream os = conn.getOutputStream()) {
                os.write(badJson.getBytes());
            }

            try (InputStream inputStream = conn.getInputStream()) {
            	//
            } catch (IOException e) {
                assertTrue(e.getMessage().startsWith(
                		"Server returned HTTP response code: 403"));
                return;
            }

            Assert.fail("Exception expected");
        } catch (IOException | NoSuchAlgorithmException | KeyManagementException e) {
            Assert.fail("Unexpected problem", e);
        }
    }
}
